@@ -1001,8 +1001,8 @@ f_win_splitmove(typval_T *argvars, typval_T *rettv)
 	size = (int)dict_get_number(d, "size");
     }
 
-    // Check if we can split the target before we bother switching windows.
-    if (text_or_buf_locked() || check_split_disallowed(targetwin) == FAIL)
+    // Check if we're allowed to continue before we bother switching windows.
+    if (text_or_buf_locked() || check_split_disallowed(wp) == FAIL)
 	return;
 
     if (curwin != targetwin)
@@ -1351,13 +1351,8 @@ switch_win_noblock(
 	switchwin->sw_curtab = curtab;
 	if (no_display)
 	{
-	    curtab->tp_firstwin = firstwin;
-	    curtab->tp_lastwin = lastwin;
-	    curtab->tp_topframe = topframe;
-	    curtab = tp;
-	    firstwin = curtab->tp_firstwin;
-	    lastwin = curtab->tp_lastwin;
-	    topframe = curtab->tp_topframe;
+	    unuse_tabpage(curtab);
+	    use_tabpage(tp);
 	}
 	else
 	    goto_tabpage_tp(tp, FALSE, FALSE);
@@ -1395,13 +1390,12 @@ restore_win_noblock(
     {
 	if (no_display)
 	{
-	    curtab->tp_firstwin = firstwin;
-	    curtab->tp_lastwin = lastwin;
-	    curtab->tp_topframe = topframe;
-	    curtab = switchwin->sw_curtab;
-	    firstwin = curtab->tp_firstwin;
-	    lastwin = curtab->tp_lastwin;
-	    topframe = curtab->tp_topframe;
+	    win_T	*old_tp_curwin = curtab->tp_curwin;
+
+	    unuse_tabpage(curtab);
+	    // Don't change the curwin of the tabpage we temporarily visited.
+	    curtab->tp_curwin = old_tp_curwin;
+	    use_tabpage(switchwin->sw_curtab);
 	}
 	else
 	    goto_tabpage_tp(switchwin->sw_curtab, FALSE, FALSE);
